@@ -1,5 +1,5 @@
 /* parser.c -- convert the command line args into an expression tree.
-   Copyright (C) 1990, 91, 92, 93, 94, 2000 Free Software Foundation, Inc.
+   Copyright (C) 1990, 91, 92, 93, 94, 2000, 2001, 2003, 2004 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -114,6 +114,7 @@ static boolean parse_printf PARAMS((char *argv[], int *arg_ptr));
 static boolean parse_prune PARAMS((char *argv[], int *arg_ptr));
 static boolean parse_regex PARAMS((char *argv[], int *arg_ptr));
 static boolean insert_regex PARAMS((char *argv[], int *arg_ptr, boolean ignore_case));
+static boolean parse_samefile PARAMS((char *argv[], int *arg_ptr));
 static boolean parse_size PARAMS((char *argv[], int *arg_ptr));
 static boolean parse_true PARAMS((char *argv[], int *arg_ptr));
 static boolean parse_type PARAMS((char *argv[], int *arg_ptr));
@@ -240,6 +241,7 @@ static struct parser_table const parse_table[] =
   {ARG_TEST,               "prune",                 parse_prune},
   {ARG_ACTION,             "quit",                  parse_quit},	/* GNU */
   {ARG_TEST,               "regex",                 parse_regex},	/* GNU */
+  {ARG_TEST,               "samefile",              parse_samefile},    /* GNU */
   {ARG_TEST,               "size",                  parse_size},
   {ARG_TEST,               "true",                  parse_true},	/* GNU */
   {ARG_TEST,               "type",                  parse_type},
@@ -296,7 +298,7 @@ find_parser (char *search_name)
 	      if (parse_table[i].type == ARG_OPTION)
 		{
 		  if ((first_nonoption_arg != NULL)
-		      && warnings )
+		      && options.warnings )
 		    {
 		      /* option which folows a non-option */
 		      error (0, 0,
@@ -352,7 +354,7 @@ parse_amin (char **argv, int *arg_ptr)
     return (false);
   if (!get_num_days (argv[*arg_ptr], &num, &c_type))
     return (false);
-  t = cur_day_start + DAYSECS - num * 60;
+  t = options.cur_day_start + DAYSECS - num * 60;
   our_pred = insert_primary (pred_amin);
   our_pred->args.info.kind = c_type;
   our_pred->args.info.negative = t < 0;
@@ -388,7 +390,7 @@ parse_anewer (char **argv, int *arg_ptr)
 
   if ((argv == NULL) || (argv[*arg_ptr] == NULL))
     return (false);
-  if ((*xstat) (argv[*arg_ptr], &stat_newer))
+  if ((*options.xstat) (argv[*arg_ptr], &stat_newer))
     error (1, errno, "%s", argv[*arg_ptr]);
   our_pred = insert_primary (pred_anewer);
   our_pred->args.time = stat_newer.st_mtime;
@@ -433,7 +435,7 @@ parse_cmin (char **argv, int *arg_ptr)
     return (false);
   if (!get_num_days (argv[*arg_ptr], &num, &c_type))
     return (false);
-  t = cur_day_start + DAYSECS - num * 60;
+  t = options.cur_day_start + DAYSECS - num * 60;
   our_pred = insert_primary (pred_cmin);
   our_pred->args.info.kind = c_type;
   our_pred->args.info.negative = t < 0;
@@ -450,7 +452,7 @@ parse_cnewer (char **argv, int *arg_ptr)
 
   if ((argv == NULL) || (argv[*arg_ptr] == NULL))
     return (false);
-  if ((*xstat) (argv[*arg_ptr], &stat_newer))
+  if ((*options.xstat) (argv[*arg_ptr], &stat_newer))
     error (1, errno, "%s", argv[*arg_ptr]);
   our_pred = insert_primary (pred_cnewer);
   our_pred->args.time = stat_newer.st_mtime;
@@ -491,17 +493,17 @@ parse_daystart (char **argv, int *arg_ptr)
   (void) argv;
   (void) arg_ptr;
 
-  if (full_days == false)
+  if (options.full_days == false)
     {
-      cur_day_start += DAYSECS;
-      local = localtime (&cur_day_start);
-      cur_day_start -= (local
+      options.cur_day_start += DAYSECS;
+      local = localtime (&options.cur_day_start);
+      options.cur_day_start -= (local
 			? (local->tm_sec + local->tm_min * 60
 			   + local->tm_hour * 3600)
-			: cur_day_start % DAYSECS);
-      full_days = true;
+			: options.cur_day_start % DAYSECS);
+      options.full_days = true;
     }
-  return (true);
+  return true;
 }
 
 static boolean
@@ -517,7 +519,7 @@ parse_delete (argv, arg_ptr)
   our_pred->side_effects = true;
   our_pred->no_default_print = true;
   /* -delete implies -depth */
-  do_dir_first = false;
+  options.do_dir_first = false;
   return (true);
 }
 
@@ -527,7 +529,7 @@ parse_depth (char **argv, int *arg_ptr)
   (void) argv;
   (void) arg_ptr;
 
-  do_dir_first = false;
+  options.do_dir_first = false;
   return (true);
 }
  
@@ -537,7 +539,7 @@ parse_d (char **argv, int *arg_ptr)
   (void) argv;
   (void) arg_ptr;
   
-  if (warnings)
+  if (options.warnings)
     {
       error (0, 0,
 	     _("warning: the -d option is deprecated; please use -depth instead, because the latter is a POSIX-compliant feature."));
@@ -703,25 +705,27 @@ parse_help (char **argv, int *arg_ptr)
   
   printf (_("\
 Usage: %s [path...] [expression]\n"), program_name);
-  puts (_("\
+  puts (_("\n\
 default path is the current directory; default expression is -print\n\
-expression may consist of:\n\
-operators (decreasing precedence; -and is implicit where no others are given):\n\
-      ( EXPR ) ! EXPR -not EXPR EXPR1 -a EXPR2 EXPR1 -and EXPR2"));
-  puts (_("      EXPR1 -o EXPR2 EXPR1 -or EXPR2 EXPR1 , EXPR2\n\
-positional options (always true): -daystart -follow\n\
-normal options (always true, specified before other expressions): -depth\n\
-      --help -maxdepth LEVELS -mindepth LEVELS -mount -noleaf --version -xdev\n\
-      -ignore_readdir_race -noignore_readdir_race\n\
-tests (N can be +N or -N or N): -amin N -anewer FILE -atime N -cmin N"));
+expression may consist of: operators, options, tests, and actions:\n"));
   puts (_("\
+operators (decreasing precedence; -and is implicit where no others are given):\n\
+      ( EXPR )   ! EXPR   -not EXPR   EXPR1 -a EXPR2   EXPR1 -and EXPR2\n\
+      EXPR1 -o EXPR2   EXPR1 -or EXPR2   EXPR1 , EXPR2\n"));
+  puts (_("\
+positional options (always true): -daystart -follow\n\
+normal options (always true, specified before other expressions):\n\
+      -depth --help -maxdepth LEVELS -mindepth LEVELS -mount -noleaf\n\
+      --version -xdev -ignore_readdir_race -noignore_readdir_race\n"));
+  puts (_("\
+tests (N can be +N or -N or N): -amin N -anewer FILE -atime N -cmin N\n\
       -cnewer FILE -ctime N -empty -false -fstype TYPE -gid N -group NAME\n\
       -ilname PATTERN -iname PATTERN -inum N -iwholename PATTERN -iregex PATTERN\n\
       -links N -lname PATTERN -mmin N -mtime N -name PATTERN -newer FILE"));
   puts (_("\
       -nouser -nogroup -path PATTERN -perm [+-]MODE -regex PATTERN\n\
       -wholename PATTERN -size N[bcwkMG] -true -type [bcdpflsD] -uid N\n\
-      -used N -user NAME -xtype [bcdpfls]"));
+      -used N -user NAME -xtype [bcdpfls]\n"));
   puts (_("\
 actions: -exec COMMAND ; -fprint FILE -fprint0 FILE -fprintf FILE FORMAT\n\
       -fls FILE -ok COMMAND ; -print -print0 -printf FORMAT -prune -ls -delete\n\
@@ -873,15 +877,15 @@ parse_maxdepth (char **argv, int *arg_ptr)
   int depth_len;
 
   if ((argv == NULL) || (argv[*arg_ptr] == NULL))
-    return (false);
+    return false;
   depth_len = strspn (argv[*arg_ptr], "0123456789");
   if ((depth_len == 0) || (argv[*arg_ptr][depth_len] != '\0'))
-    return (false);
-  maxdepth = atoi (argv[*arg_ptr]);
-  if (maxdepth < 0)
-    return (false);
+    return false;
+  options.maxdepth = atoi (argv[*arg_ptr]);
+  if (options.maxdepth < 0)
+    return false;
   (*arg_ptr)++;
-  return (true);
+  return true;
 }
 
 static boolean
@@ -890,15 +894,15 @@ parse_mindepth (char **argv, int *arg_ptr)
   int depth_len;
 
   if ((argv == NULL) || (argv[*arg_ptr] == NULL))
-    return (false);
+    return false;
   depth_len = strspn (argv[*arg_ptr], "0123456789");
   if ((depth_len == 0) || (argv[*arg_ptr][depth_len] != '\0'))
-    return (false);
-  mindepth = atoi (argv[*arg_ptr]);
-  if (mindepth < 0)
-    return (false);
+    return false;
+  options.mindepth = atoi (argv[*arg_ptr]);
+  if (options.mindepth < 0)
+    return false;
   (*arg_ptr)++;
-  return (true);
+  return true;
 }
 
 static boolean
@@ -913,7 +917,7 @@ parse_mmin (char **argv, int *arg_ptr)
     return (false);
   if (!get_num_days (argv[*arg_ptr], &num, &c_type))
     return (false);
-  t = cur_day_start + DAYSECS - num * 60;
+  t = options.cur_day_start + DAYSECS - num * 60;
   our_pred = insert_primary (pred_mmin);
   our_pred->args.info.kind = c_type;
   our_pred->args.info.negative = t < 0;
@@ -975,7 +979,7 @@ parse_newer (char **argv, int *arg_ptr)
   
   if ((argv == NULL) || (argv[*arg_ptr] == NULL))
     return (false);
-  if ((*xstat) (argv[*arg_ptr], &stat_newer))
+  if ((*options.xstat) (argv[*arg_ptr], &stat_newer))
     error (1, errno, "%s", argv[*arg_ptr]);
   our_pred = insert_primary (pred_newer);
   our_pred->args.time = stat_newer.st_mtime;
@@ -989,7 +993,7 @@ parse_noleaf (char **argv, int *arg_ptr)
   (void) &argv;
   (void) &arg_ptr;
   
-  no_leaf_check = true;
+  options.no_leaf_check = true;
   return true;
 }
 
@@ -1087,7 +1091,7 @@ parse_nowarn (char **argv, int *arg_ptr)
   (void) argv;
   (void) arg_ptr;
   
-  warnings = false;
+  options.warnings = false;
   return true;;
 }
 
@@ -1396,6 +1400,27 @@ parse_size (char **argv, int *arg_ptr)
   return (true);
 }
 
+
+static boolean
+parse_samefile (char **argv, int *arg_ptr)
+{
+  struct predicate *our_pred;
+  struct stat st;
+  
+  if ((argv == NULL) || (argv[*arg_ptr] == NULL))
+    return (false);
+  if ((*options.xstat) (argv[*arg_ptr], &st))
+    error (1, errno, "%s", argv[*arg_ptr]);
+  
+  our_pred = insert_primary (pred_samefile);
+  our_pred->args.fileid.ino = st.st_ino;
+  our_pred->args.fileid.dev = st.st_dev;
+  our_pred->need_stat = true;
+  (*arg_ptr)++;
+  return (true);
+}
+
+
 static boolean
 parse_true (char **argv, int *arg_ptr)
 {
@@ -1486,7 +1511,7 @@ parse_xdev (char **argv, int *arg_ptr)
 {
   (void) argv;
   (void) arg_ptr;
-  stay_on_filesystem = true;
+  options.stay_on_filesystem = true;
   return true;
 }
 
@@ -1495,7 +1520,7 @@ parse_ignore_race (char **argv, int *arg_ptr)
 {
   (void) argv;
   (void) arg_ptr;
-  ignore_readdir_race = true;
+  options.ignore_readdir_race = true;
   return true;
 }
 
@@ -1504,7 +1529,7 @@ parse_noignore_race (char **argv, int *arg_ptr)
 {
   (void) argv;
   (void) arg_ptr;
-  ignore_readdir_race = false;
+  options.ignore_readdir_race = false;
   return true;
 }
 
@@ -1513,7 +1538,7 @@ parse_warn (char **argv, int *arg_ptr)
 {
   (void) argv;
   (void) arg_ptr;
-  warnings = true;
+  options.warnings = true;
   return true;
 }
 
@@ -1910,16 +1935,16 @@ insert_time (char **argv, int *arg_ptr, PFB pred)
     return (false);
 
   /* Figure out the timestamp value we are looking for. */
-  t = ( cur_day_start - num_days * DAYSECS
+  t = ( options.cur_day_start - num_days * DAYSECS
 		   + ((c_type == COMP_GT) ? DAYSECS - 1 : 0));
-#if 1
+
   if (1)
     {
       /* We introduce a scope in which 'val' can be declared, for the 
        * benefit of compilers that are really C89 compilers
        * which support intmax_t because config.h #defines it
        */
-      intmax_t val = ( (intmax_t)cur_day_start - num_days * DAYSECS
+      intmax_t val = ( (intmax_t)options.cur_day_start - num_days * DAYSECS
 		       + ((c_type == COMP_GT) ? DAYSECS - 1 : 0));
       t = val;
       
@@ -1929,7 +1954,6 @@ insert_time (char **argv, int *arg_ptr, PFB pred)
 	  error (1, 0, "arithmetic overflow while converting %s days to a number of seconds", argv[*arg_ptr]);
 	}
     }
-#endif
   
   our_pred = insert_primary (pred);
   our_pred->args.info.kind = c_type;
