@@ -1,5 +1,5 @@
 /* pred.c -- execute the expression tree.
-   Copyright (C) 1990, 91, 92, 93, 94, 2000, 2003 Free Software Foundation, Inc.
+   Copyright (C) 1990, 91, 92, 93, 94, 2000, 2003, 2004 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -194,6 +194,7 @@ struct pred_assoc pred_table[] =
   {pred_print0, "print0  "},
   {pred_prune, "prune   "},
   {pred_regex, "regex   "},
+  {pred_samefile,"samefile "},
   {pred_size, "size    "},
   {pred_true, "true    "},
   {pred_type, "type    "},
@@ -295,13 +296,13 @@ pred_and (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
       /* Check whether we need a stat here. */
       if (pred_ptr->need_stat)
 	{
-	  if (!have_stat && (*xstat) (rel_pathname, stat_buf) != 0)
+	  if (!state.have_stat && (*options.xstat) (state.rel_pathname, stat_buf) != 0)
 	    {
 	      error (0, errno, "%s", pathname);
-	      exit_status = 1;
+	      state.exit_status = 1;
 	      return (false);
 	    }
-	  have_stat = true;
+	  state.have_stat = true;
 	}
       return ((*pred_ptr->pred_right->pred_func) (pathname, stat_buf,
 						  pred_ptr->pred_right));
@@ -364,13 +365,13 @@ pred_comma (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
   /* Check whether we need a stat here. */
   if (pred_ptr->need_stat)
     {
-      if (!have_stat && (*xstat) (rel_pathname, stat_buf) != 0)
+      if (!state.have_stat && (*options.xstat) (state.rel_pathname, stat_buf) != 0)
 	{
 	  error (0, errno, "%s", pathname);
-	  exit_status = 1;
+	  state.exit_status = 1;
 	  return (false);
 	}
-      have_stat = true;
+      state.have_stat = true;
     }
   return ((*pred_ptr->pred_right->pred_func) (pathname, stat_buf,
 					      pred_ptr->pred_right));
@@ -388,9 +389,9 @@ pred_delete (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pred_ptr;
   (void) stat_buf;
-  if (strcmp (rel_pathname, "."))
+  if (strcmp (state.rel_pathname, "."))
     {
-      if (0 != remove (rel_pathname))
+      if (0 != remove (state.rel_pathname))
 	{
 	  error (0, errno, "cannot delete %s", pathname);
 	  return false;
@@ -418,12 +419,12 @@ pred_empty (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
       boolean empty = true;
 
       errno = 0;
-      d = opendir (rel_pathname);
+      d = opendir (state.rel_pathname);
       if (d == NULL)
 	{
 	  error (0, errno, "%s", pathname);
-	  exit_status = 1;
-	  return (false);
+	  state.exit_status = 1;
+	  return false;
 	}
       for (dp = readdir (d); dp; dp = readdir (d))
 	{
@@ -438,8 +439,8 @@ pred_empty (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
       if (CLOSEDIR (d))
 	{
 	  error (0, errno, "%s", pathname);
-	  exit_status = 1;
-	  return (false);
+	  state.exit_status = 1;
+	  return false;
 	}
       return (empty);
     }
@@ -504,8 +505,8 @@ pred_false (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 boolean
 pred_fls (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
-  list_file (pathname, rel_pathname, stat_buf, start_time,
-	     output_block_size, pred_ptr->args.stream);
+  list_file (pathname, state.rel_pathname, stat_buf, options.start_time,
+	     options.output_block_size, pred_ptr->args.stream);
   return (true);
 }
 
@@ -587,7 +588,7 @@ pred_fprintf (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 	  fprintf (fp, segment->text, ctime_format (stat_buf->st_ctime));
 	  break;
 	case 'd':		/* depth in search tree */
-	  fprintf (fp, segment->text, curdepth);
+	  fprintf (fp, segment->text, state.curdepth);
 	  break;
 	case 'D':		/* Device on which file exists (stat.st_dev) */
 	  fprintf (fp, segment->text, 
@@ -598,8 +599,7 @@ pred_fprintf (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 	  fprintf (fp, segment->text, base_name (pathname));
 	  break;
 	case 'F':		/* filesystem type */
-	  fprintf (fp, segment->text,
-		   filesystem_type (pathname, rel_pathname, stat_buf));
+	  fprintf (fp, segment->text, filesystem_type (stat_buf));
 	  break;
 	case 'g':		/* group name */
 	  {
@@ -634,11 +634,11 @@ pred_fprintf (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 	  }
 	case 'H':		/* ARGV element file was found under */
 	  {
-	    char cc = pathname[path_length];
+	    char cc = pathname[state.path_length];
 
-	    pathname[path_length] = '\0';
+	    pathname[state.path_length] = '\0';
 	    fprintf (fp, segment->text, pathname);
-	    pathname[path_length] = cc;
+	    pathname[state.path_length] = cc;
 	    break;
 	  }
 	case 'i':		/* inode number */
@@ -660,9 +660,9 @@ pred_fprintf (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 
 	    if (S_ISLNK (stat_buf->st_mode))
 	      {
-		linkname = get_link_name (pathname, rel_pathname);
+		linkname = get_link_name (pathname, state.rel_pathname);
 		if (linkname == 0)
-		  exit_status = 1;
+		  state.exit_status = 1;
 	      }
 	    if (linkname)
 	      {
@@ -725,9 +725,9 @@ pred_fprintf (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 	  fprintf (fp, segment->text, pathname);
 	  break;
 	case 'P':		/* pathname with ARGV element stripped */
-	  if (curdepth)
+	  if (state.curdepth > 0)
 	    {
-	      cp = pathname + path_length;
+	      cp = pathname + state.path_length;
 	      if (*cp == '/')
 		/* Move past the slash between the ARGV element
 		   and the rest of the pathname.  But if the ARGV element
@@ -775,8 +775,8 @@ pred_fprintf (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 	    {
 	      struct stat sbuf;
 	      int (*ystat) ();
-	      ystat = xstat == lstat ? stat : lstat;
-	      if ((*ystat) (rel_pathname, &sbuf) != 0)
+	      ystat = options.xstat == lstat ? stat : lstat;
+	      if ((*ystat) (state.rel_pathname, &sbuf) != 0)
 	      {
 		if ( errno == ENOENT ) {
 		  fprintf (fp, segment->text, "N");
@@ -821,10 +821,12 @@ pred_fprintf (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 boolean
 pred_fstype (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
-  if (strcmp (filesystem_type (pathname, rel_pathname, stat_buf),
-	      pred_ptr->args.str) == 0)
-    return (true);
-  return (false);
+  (void) pathname;
+  
+  if (strcmp (filesystem_type (stat_buf), pred_ptr->args.str) == 0)
+    return true;
+  else
+    return false;
 }
 
 boolean
@@ -952,7 +954,7 @@ insert_lname (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr,
 #ifdef S_ISLNK
   if (S_ISLNK (stat_buf->st_mode))
     {
-      char *linkname = get_link_name (pathname, rel_pathname);
+      char *linkname = get_link_name (pathname, state.rel_pathname);
       if (linkname)
 	{
 	  if (fnmatch (pred_ptr->args.str, linkname,
@@ -970,8 +972,8 @@ pred_ls (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   (void) pred_ptr;
   
-  list_file (pathname, rel_pathname, stat_buf, start_time,
-	     output_block_size, stdout);
+  list_file (pathname, state.rel_pathname, stat_buf, options.start_time,
+	     options.output_block_size, stdout);
   return (true);
 }
 
@@ -1011,13 +1013,13 @@ pred_negate (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
   /* Check whether we need a stat here. */
   if (pred_ptr->need_stat)
     {
-      if (!have_stat && (*xstat) (rel_pathname, stat_buf) != 0)
+      if (!state.have_stat && (*options.xstat) (state.rel_pathname, stat_buf) != 0)
 	{
 	  error (0, errno, "%s", pathname);
-	  exit_status = 1;
-	  return (false);
+	  state.exit_status = 1;
+	  return false;
 	}
-      have_stat = true;
+      state.have_stat = true;
     }
   return (!(*pred_ptr->pred_right->pred_func) (pathname, stat_buf,
 					      pred_ptr->pred_right));
@@ -1102,19 +1104,19 @@ pred_or (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
       /* Check whether we need a stat here. */
       if (pred_ptr->need_stat)
 	{
-	  if (!have_stat && (*xstat) (rel_pathname, stat_buf) != 0)
+	  if (!state.have_stat && (*options.xstat) (state.rel_pathname, stat_buf) != 0)
 	    {
 	      error (0, errno, "%s", pathname);
-	      exit_status = 1;
-	      return (false);
+	      state.exit_status = 1;
+	      return false;
 	    }
-	  have_stat = true;
+	  state.have_stat = true;
 	}
       return ((*pred_ptr->pred_right->pred_func) (pathname, stat_buf,
 						  pred_ptr->pred_right));
     }
   else
-    return (true);
+    return true;
 }
 
 boolean
@@ -1175,8 +1177,8 @@ pred_prune (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
   (void) pathname;
   (void) stat_buf;
   (void) pred_ptr;
-  stop_at_current_level = true;
-  return (do_dir_first);	/* This is what SunOS find seems to do. */
+  state.stop_at_current_level = true;
+  return (options.do_dir_first); /* This is what SunOS find seems to do. */
 }
 
 boolean
@@ -1225,6 +1227,21 @@ pred_size (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
   return (false);
 }
 
+boolean
+pred_samefile (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+{
+  /* Potential optimisation: because of the loop protection, we 
+   * always know the device of the current directory, hence the 
+   * device number of the current filesystem.  If -L is not in 
+   * effect, and the device number of the file we're looking for 
+   * is not the same as the device number of the current directory,
+   * this predicate cannot return true.   Hence there would be no 
+   * need to stat the file.
+   */
+  return stat_buf->st_ino == pred_ptr->args.fileid.ino
+    &&   stat_buf->st_dev == pred_ptr->args.fileid.dev;
+}
+
 boolean
 pred_true (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
@@ -1318,7 +1335,7 @@ pred_xtype (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
   struct stat sbuf;		/* local copy, not stat_buf because we're using a different stat method */
   int (*ystat) (const char*, struct stat *p);
 
-  switch (symlink_handling)
+  switch (options.symlink_handling)
     {
     case SYMLINK_ALWAYS_DEREF:
       ystat = optionl_stat;
@@ -1327,7 +1344,7 @@ pred_xtype (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
       ystat = optionp_stat;
     }
   
-  if ((*ystat) (rel_pathname, &sbuf) != 0)
+  if ((*ystat) (state.rel_pathname, &sbuf) != 0)
     {
       if (ystat == optionl_stat && errno == ENOENT)
 	{
@@ -1340,7 +1357,7 @@ pred_xtype (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
       else
 	{
 	  error (0, errno, "%s", pathname);
-	  exit_status = 1;
+	  state.exit_status = 1;
 	}
       return false;
     }
@@ -1412,14 +1429,14 @@ launch (struct predicate *pred_ptr)
     if (errno != EINTR)
       {
 	error (0, errno, _("error waiting for %s"), execp->vec[0]);
-	exit_status = 1;
+	state.exit_status = 1;
 	return false;
       }
   if (WIFSIGNALED (status))
     {
       error (0, 0, _("%s terminated by signal %d"),
 	     execp->vec[0], WTERMSIG (status));
-      exit_status = 1;
+      state.exit_status = 1;
       return (false);
     }
   return (!WEXITSTATUS (status));
@@ -1520,27 +1537,25 @@ prec_name (prec)
    INDENT is the number of levels to indent the left margin. */
 
 void
-print_tree (node, indent)
-     struct predicate *node;
-     int indent;
+print_tree (FILE *fp, struct predicate *node, int indent)
 {
   int i;
 
   if (node == NULL)
     return;
   for (i = 0; i < indent; i++)
-    printf ("    ");
-  printf ("pred = %s type = %s prec = %s addr = %x\n",
+    fprintf (fp, "    ");
+  fprintf (fp, "pred = %s type = %s prec = %s addr = %p\n",
 	  find_pred_name (node->pred_func),
 	  type_name (node->p_type), prec_name (node->p_prec), node);
   for (i = 0; i < indent; i++)
-    printf ("    ");
-  printf (_("left:\n"));
-  print_tree (node->pred_left, indent + 1);
+    fprintf (fp, "    ");
+  fprintf (fp, _("left:\n"));
+  print_tree (fp, node->pred_left, indent + 1);
   for (i = 0; i < indent; i++)
-    printf ("    ");
-  printf (_("right:\n"));
-  print_tree (node->pred_right, indent + 1);
+    fprintf (fp, "    ");
+  fprintf (fp, _("right:\n"));
+  print_tree (fp, node->pred_right, indent + 1);
 }
 
 /* Copy STR into BUF and trim blanks from the end of BUF.
@@ -1566,8 +1581,7 @@ blank_rtrim (str, buf)
 /* Print out the predicate list starting at NODE. */
 
 void
-print_list (node)
-     struct predicate *node;
+print_list (FILE *fp, struct predicate *node)
 {
   struct predicate *cur;
   char name[256];
@@ -1575,9 +1589,60 @@ print_list (node)
   cur = node;
   while (cur != NULL)
     {
-      printf ("%s ", blank_rtrim (find_pred_name (cur->pred_func), name));
+      fprintf (fp, "%s ", blank_rtrim (find_pred_name (cur->pred_func), name));
       cur = cur->pred_next;
     }
-  printf ("\n");
+  fprintf (fp, "\n");
 }
+
+
+/* Print out the predicate list starting at NODE. */
+
+
+static void
+print_parenthesised(FILE *fp, struct predicate *node)
+{
+  int parens = 0;
+
+  if (node)
+    {
+      if ( ( (node->pred_func == pred_or)
+	     || (node->pred_func == pred_and) )
+	  && node->pred_left == NULL)
+	{
+	  /* We print "<nothing> or  X" as just "X"
+	   * We print "<nothing> and X" as just "X"
+	   */
+	  print_parenthesised(fp, node->pred_right);
+	}
+      else
+	{
+	  if (node->pred_left || node->pred_right)
+	    parens = 1;
+
+	  if (parens)
+	    fprintf(fp, "%s", " ( ");
+	  print_optlist(fp, node);
+	  if (parens)
+	    fprintf(fp, "%s", " ) ");
+	}
+    }
+}
+
+void
+print_optlist (FILE *fp, struct predicate *p)
+{
+  char name[256];
+
+  if (p)
+    {
+      print_parenthesised(fp, p->pred_left);
+      fprintf (fp,
+	       "%s%s ",
+	       p->need_stat ? _("[stat called here] ") : "",
+	       blank_rtrim (find_pred_name (p->pred_func), name));
+      print_parenthesised(fp, p->pred_right);
+    }
+}
+
 #endif	/* DEBUG */
